{
  Based on TListDataset by DrBob

  Modified by Roland Potter - Woodstock Computer Solutions :
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.

    Internal list is now a list of Indexes into an external list
    (Base class only knows External items as an integer).

    Having the 2 lists enables :
      1 : non "TList" based External Lists
      2 : Sorting while maintaining original list order
      3 : Filtering

    SubClasses must provide :
      ListInsert    - Add to external list and return new index.
      ListDelete    - Delete from external list.
      ListCount     - Count of external list.
      ListGetObject - The object pointed to by the external list.
      ListSetObject - To replace object after editing.
      ListGetField  - Get the objects fields into a TField Buffer.
      ListSetField  - Set the Object fields from a TField Buffer.
      ListCreateElement - Create an UNLISTED element for editing.
      ListFreeElement   - Free an UNLISTED element after a cancel.

    If Blobs used, must provide :
      ListGetBlob   - Get the objects blob field into a Buffer, or get size
      ListSetBlob   - Set the Objects blob field from a Buffer
    Optionally can Provide :
      ListSetupFieldDefs - Use FieldDefs.Add to add fields if no persistent
      ListSetupFields - fiddle with automated fields

    Also can provide compare routines for sorting.

}

unit wdcDset;

interface

uses
  DB, Classes, SysUtils, Windows, Forms;

type
  TWcsMemDataSetCompare = function(Object1, Object2: TObject): Integer of object;
  PLDSRecInfo = ^TLDSRecInfo;
  TLDSRecInfo = record
    IntIdx: Longint;
    Bookmark: Pointer; // Actually stores Longint
    BookmarkFlag: TBookmarkFlag;
  end;
  TWcsMemDataSet = class (TDataSet)
  private
    fSortList: PPointerList;
    fOnSortCompare: TWcsMemDataSetCompare;
    fQSLevel: integer;
    fReadOnly: Boolean;
    procedure SetExtList(const Value: TObject);
    procedure QuickSort(L, R: Integer;
      const SCompare: TWcsMemDataSetCompare);
    function GetRecordCheck(GetMode: TGetMode): Integer;
    procedure FillBuffer(const Buffer: PChar;
      const InternalIndex: integer;
      const BkmkFlag: TBookmarkFlag);
    function FilteredRecordIndex(const TestIdx: integer): integer;
    procedure IndexReset(const IdxFrom, IdxTo, IdxChange: Integer);
    procedure SetReadOnly(const Value: Boolean);
  protected
    // Subclasses may find these useful so protected not private
    // record data and status
    fIsTableOpen: Boolean;
    fIdxList: TList;
   // To incorporate full record buffering...
    fRecBufSize: Integer; // 0 FieldBuffer Size
    fCalcBufSize,
    fBlobBufSize,
    fRecordSize: Integer; // {FieldBuffer Size +} TLDSRecInfo
    fCalcBufOffset,
    fBlobBufOffset: Integer;
    fCurrent: Integer;
    fExtList : TObject;
    fEditBuff : PChar;
    fEditObject : TObject;
    // for true buffering
    fDataOffsets : array of Integer;
    fDataSizes   : array of Integer;

    fFilterBuffer: PChar;

    // dataset virtual methods
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalInsert; override;
    procedure InternalEdit; override;
    procedure InternalCancel; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetRecordCount: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    function GetRecNo: Integer; override;
    function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
    procedure InternalInitFieldDefs; override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;

    // Grids/datalink buffers mess up our current record
    // indicator with buffering.
    procedure DoAfterScroll; override;

    // This always just clears the list and reads the indices into the list
    procedure ReadListData; virtual;

    // This does the initial checking for a valid RecBuffer and List Index
    // Used by GetFieldData and SetFieldData
    function GetActiveRecBuffer: PChar;  virtual;
    function RecInfoFromBuffer(Buffer: PChar) : PLDSRecInfo;  virtual;
    function RecObjFromBuffer(Buffer: PChar): TObject;

    //========================================================================
    // for specific subclasses to override...
    function ListCount: Integer; virtual; abstract;
    function ListGetObject(index : integer) : TObject; virtual; abstract;
    // Internal indexing in the External list may reposition on change of data
    // Descendants must free the existing object if any.
    procedure ListSetObject(NewObj: TObject; var NewIdx: Longint); virtual; abstract;
    // Move data from TField buffers back to objects
    procedure ListSetField(ListObject : TObject;
                           Field: TField; FieldBuf: Pointer); virtual; abstract;
    // Move data from Objects to TField style buffers - Return False if Null
    // List Index is passed for convenience - will be -1 for new recs
    function ListGetField(ListObject : TObject; ListIndex: integer;
                          Field: TField; FieldBuf: Pointer): Boolean; virtual; abstract;
    // Move data direct from Blob to Buffer
    function ListGetBlob(const ListObject : TObject;
                         const Field: TField;
                         const Buffer : Pointer;
                         var   BufLen : integer): Boolean; virtual; abstract;
    // Move data from Objects to TField style buffers - Return False if Null
    procedure ListSetBlob(const ListObject : TObject;
                         const Field: TField;
                         const Buffer: Pointer;
                         const BufLen: Integer); virtual; abstract;
    function ListCreateElement(const Element: TObject): TObject; virtual; abstract;
    procedure ListFreeElement(const Element: TObject); virtual; abstract;

    // The underlying list can position where it wants, and return the index
    // The index passed in, is the position of the insert in the dataset.
    procedure ListInsert(const NewObj : TObject; var NewIdx: integer); virtual; abstract;
    procedure ListDelete(const DelIdx : integer); virtual; abstract;

    // This can be overriden if auto fields required else leave empty
    // Use FieldDefs.Clear and then AddFieldDefs etc
    procedure ListSetupFieldDefs; virtual;
    // This allows manipulation of TFields array after it is setup
    procedure ListSetupFields; virtual;

    // For List objects that must go off to a file or other media
    // to fill themselves
    procedure ListDataFetch; virtual;

    // For Indexed DataSets for an initial Sort before you know anything's happened
    procedure ListDataRead; virtual;

    //========================================================================
  public
    constructor Create (Owner: TComponent); override;
    destructor Destroy; override;

    // These check the ListIndex and call the new abstracts above
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;

    procedure SortByFunc(const SortFunc : TWcsMemDataSetCompare);
   // For sorting using OnSortCompare
    procedure Sort;
    procedure ListIndexSort;

    property ExtList : TObject read fExtList write SetExtList;
  published
    property ReadOnly: Boolean read fReadOnly write SetReadOnly default False;
    // redeclared dataset properties
    property Filtered;

    property Active;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    // For sorting
    property OnSortCompare : TWcsMemDataSetCompare read fOnSortCompare write fOnSortCompare;
  end;

  // This is a subclass where the external list IS a TList
  TWcsMemListDataSet = class(TWcsMemDataSet)
  private
    function getDBList: TList;
    procedure SetDBList(const Value: TList);
  protected
    function ListCount: Integer; override;
    function ListGetObject(index : integer) : TObject; override;
    procedure ListSetObject(NewObj: TObject; var NewIdx: Longint); override;
    procedure ListFreeElement(const Element: TObject); override;
    procedure ListInsert(const NewObj : TObject; var NewIdx: integer); override;
    procedure ListDelete(const DelIdx : integer); override;
  { specific subclasses still Need to add at least.....
    procedure ListSetupFieldDefs; override;
    function ListCreateElement(const Element: TObject): TObject; override;
    procedure ListSetField(ListObject : TObject;
                           Field: TField; FieldBuf: Pointer); override;
    function ListGetField(ListObject : TObject; ListIndex: integer;
                          Field: TField; FieldBuf: Pointer): Boolean; override;
    // Only if Blobs used....
    function ListGetBlob(const ListObject : TObject; const Field: TField;
      const Buffer : Pointer; var BufLen : integer): Boolean; override;
    procedure ListSetBlob(const ListObject : TObject; const Field: TField;
      const Buffer: Pointer; const BufLen: Integer); override;
  }
  public
    property DBList : TList read getDBList write SetDBList;
  end;

  TArticleItem = class(TObject)
  private
    fArticleID: DWord;
    fArticleTitle: String;
    fArticleText: String;  // potentially LONG
  public
    property ArticleID: DWord read fArticleID;
    property ArticleTitle: String read fArticleTitle write fArticleTitle;
    property ArticleText: String read fArticleText write fArticleText;
  end;
  // This dataset shows usage of the above for a very simple
  // database of articles, and adds a field to display the index
  // It probably has no earthly use - but makes a simple example!
  TWcsArticleDataSet = class(TWcsMemListDataSet)
  private
    fLastId: DWord;
    fArticles : TList;
    procedure ClearArticles;
  protected
    procedure ListSetupFieldDefs; override;
    function ListCreateElement(const Element: TObject): TObject; override;
    procedure ListSetField(ListObject : TObject;
                           Field: TField; FieldBuf: Pointer); override;
    function ListGetField(ListObject : TObject; ListIndex: integer;
                          Field: TField; FieldBuf: Pointer): Boolean; override;
    function ListGetBlob(const ListObject : TObject; const Field: TField;
      const Buffer : Pointer; var BufLen : integer): Boolean; override;
    procedure ListSetBlob(const ListObject : TObject; const Field: TField;
      const Buffer: Pointer; const BufLen: Integer); override;
    // Only overridden to set the ID field
    procedure ListInsert(const NewObj : TObject; var NewIdx: integer); override;

    function DBIdCompare(Object1, Object2 : TObject): Integer;
    function DBTitleCompare(Object1, Object2 : TObject): Integer;
  public
    constructor Create (Owner: TComponent); override;
    destructor Destroy; override;
    procedure AlphaSort;
    procedure IDSort;
    procedure SaveToStream(const Stream: TStream);
    procedure LoadFromStream(const Stream: TStream);
  end;

  TWcsMemDataSetBlobStream = class(TMemoryStream)
  private
    fMode : TBlobStreamMode;
    fField : TField;
    fObject : TObject;
  public
    constructor Create(AField : TField; AMode:TBlobStreamMode);
    destructor Destroy; override;
  end;
const
  NewRecordFlag : integer = -1;

implementation
uses
  dbConsts;

{ TWcsMemDataSet }

function TWcsMemDataSet.AllocRecordBuffer: PChar;
begin
  Result := StrAlloc(fRecordSize);
end;

procedure TWcsMemDataSet.InternalInitRecord(Buffer: PChar);
begin
  FillChar(Buffer^, FRecordSize, 0);
end;

procedure TWcsMemDataSet.FreeRecordBuffer (var Buffer: PChar);
begin
  StrDispose(Buffer);
end;

procedure TWcsMemDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PPointer(Data)^ := PLDSRecInfo(Buffer+fRecBufSize).Bookmark;
end;

function TWcsMemDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PLDSRecInfo(Buffer+fRecBufSize).BookmarkFlag;
end;

procedure TWcsMemDataSet.InternalGotoBookmark(Bookmark: Pointer);
var
  P: Pointer;
begin
  if (Bookmark <> nil) then
    fCurrent := fIdxList.IndexOf(PPointer(Bookmark)^)+1;
end;

function TWcsMemDataSet.GetRecNo: Integer;
begin
  Result := fCurrent;
end;

procedure TWcsMemDataSet.FillBuffer(const Buffer: PChar;
  const InternalIndex: integer;
  const BkmkFlag: TBookmarkFlag);
begin
  try
//    ClearBlobBuf(Buffer);
    GetCalcFields(@Buffer[fCalcBufOffset]);
  finally
  end;
  with PLDSRecInfo(Buffer+fRecBufSize)^ do
  begin
    BookmarkFlag := BkmkFlag;
    IntIdx       := InternalIndex;
    Bookmark     := fIdxList[InternalIndex-1];
  end;
end;

function TWcsMemDataSet.FilteredRecordIndex(const TestIdx: integer): integer;
var
  FilterRes: Boolean;
  SaveState: TDatasetState;
begin
  Result := TestIdx;
  if (TestIdx > 0) and Filtered and Assigned(OnFilterRecord) then
  begin
    SaveState := SetTempState(dsFilter);
    try
      FilterRes:= True;
      FillBuffer(fFilterBuffer, TestIdx, bfCurrent);
      OnFilterRecord(Self, FilterRes);
      if not FilterRes then
        Result := 0;
//      ClearBlobBuf(fFilterBuffer);
      RestoreState(SaveState);
    except
      RestoreState(SaveState);
      Result := 0;
//      ClearBlobBuf(fFilterBuffer);
      raise;
    end;
  end;
end;

function TWcsMemDataSet.GetRecordCheck(GetMode: TGetMode): integer;
var
  TestIdx : integer;
begin
  Result := 0;
  TestIdx:= fCurrent;
  case GetMode of
    gmNext:
      while (Result = 0) and (TestIdx < fIdxList.Count) do
      begin
        Inc (TestIdx);
        Result := FilteredRecordIndex(TestIdx);
      end;
    gmPrior:
      while (Result = 0) and (TestIdx > 1) do
      begin
        Dec (TestIdx);
        Result := FilteredRecordIndex(TestIdx);
      end;
    gmCurrent: // check if empty
      if (TestIdx <= fIdxList.Count) then
        Result := FilteredRecordIndex(TestIdx);
  end;
end;

function TWcsMemDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
  i : integer;
  FieldBuf : Pointer;
  fRecordObj : TObject;
  RecFound : Boolean;
begin
  Result := grOK; // default
  fCurrent := GetRecordCheck(GetMode);
  if fCurrent = 0 then
    case GetMode of
      gmNext   : Result := grEOF;
      gmPrior  : Result := grBOF;
      gmCurrent: Result := grEOF;
    end;

  if Result = grOK then // read the data - not really - just put in the indexes!
    FillBuffer(Buffer, fCurrent, bfCurrent);
end;

function TWcsMemDataSet.GetRecordCount: Integer;
begin
  Result := FIdxList.Count;
end;

function TWcsMemDataSet.GetRecordSize: Word;
begin
  Result := 0;
end;

procedure TWcsMemDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  if Append then SetBookmarkFlag(Buffer, bfEOF);
end;

procedure TWcsMemDataSet.InternalClose;
begin
  // disconnect and destroy field objects
  BindFields (False);
  if DefaultFields then
    DestroyFields;
  // closed
  FIsTableOpen := False;
  FreeMem(fFilterBuffer);
end;

procedure TWcsMemDataSet.IndexReset(const IdxFrom, IdxTo, IdxChange: Longint);
var
  i, ExtIdx : LongInt;
  ExtIdxP : PLongint;
begin
// This is the price paid for listing the external index
// rather than the object. Made up for by not having to find the
// object at the other end.
  for i := 0 to fIdxList.Count - 1 do
  begin
    ExtIdx := Longint(fIdxList[i]);
    if (ExtIdx >= IdxFrom) and (ExtIdx <= IdxTo) then
      fIdxList[i] := Pointer(ExtIdx + IdxChange);
  end;
// Also need to adjust all the (used) buffered records
  for i := 0 to BufferCount-1 do
    if Buffers[i] <> Nil then
    with PLDSRecInfo(Buffers[i]+fRecBufSize)^ do
    if IntIdx > 0 then
    begin
      ExtIdxP := PLongint(@Bookmark);
      if (ExtIdxP^ >= IdxFrom) and (ExtIdxP^ <= IdxTo) then
        ExtIdxP^ := ExtIdxP^ + IdxChange;
    end;
end;

procedure TWcsMemDataSet.InternalDelete;
var
  ExtIdx : integer;
begin
{
  if fNewRecord then
  begin
    InternalCancel;
    Exit;
  end;
}
  ExtIdx := Longint(fIdxList[fCurrent-1]);
  ListDelete(ExtIdx);
  fIdxList.Delete(fCurrent-1);
  if fCurrent > fIdxList.Count then dec(fCurrent);
  IndexReset(ExtIdx+1, ListCount, -1);
end;

procedure TWcsMemDataSet.InternalFirst;
begin
  fCurrent := 0;
end;

procedure TWcsMemDataSet.InternalHandleException;
begin
  Application.HandleException(Self);
end;

procedure TWcsMemDataSet.InternalEdit;
begin
  fEditObject := ListCreateElement(RecObjFromBuffer(GetActiveRecBuffer));
  fEditBuff   := GetActiveRecBuffer;
end;

procedure TWcsMemDataSet.InternalInsert;
begin
// Add on the end - post may move if insert rather than append
  fEditObject := ListCreateElement(Nil);
// Do NOT put in any lists yet
  fEditBuff   := GetActiveRecBuffer;
  with PLDSRecInfo(fEditBuff+fRecBufSize)^ do
  begin
    if fCurrent = 0 then
      BookMark := Nil
    else
      BookMark := fIdxList[fCurrent-1];
    IntIdx := fCurrent;
  end;
end;

procedure TWcsMemDataSet.InternalCancel;
begin
  try
    ListFreeElement(fEditObject);
  finally
    fEditObject := Nil;
  end;
end;

procedure TWcsMemDataSet.InternalLast;
begin
  fCurrent := FIdxList.Count+1;
end;

procedure TWcsMemDataSet.InternalOpen;
var
  TotFieldBuff, i : integer;
begin
  // initialize field definitions and create fields
  InternalInitFieldDefs;
  if DefaultFields then
    CreateFields;
  BindFields (True);
  ListSetupFields;

  TotFieldBuff := 0;
  SetLength(fDataOffsets, FieldCount);
  SetLength(fDataSizes, FieldCount);
  for i := 0 to FieldCount - 1 do
  begin
    if Fields[i].Calculated then // we are NOT handling these yet!
    begin
      fDataSizes[i]    := 0;
      fDataOffsets[i]  := -1;
    end else begin
      fDataSizes[i]   := Fields[i].DataSize; // blobs will return zero
      fDataOffsets[i] := TotFieldBuff;
      inc(TotFieldBuff, fDataSizes[i]+1); // + NullFlag!
    end;
  end;
  // Add full record buffering by uncommenting and check for other
  // comments regarding Full Buffering
  // fRecBufSize := TotFieldBuff;

  // Set up a list of objects
  ListDataFetch;
  ReadListData;
  ListDataRead;

  // initialize
  fCalcBufOffset := sizeof (TLDSRecInfo);
  fCalcBufSize   := CalcFieldsSize;
  fBlobBufOffset := fCalcBufOffset + fCalcBufSize;
  fBlobBufSize   := Sizeof(Pointer) * BlobFieldCount;
  fRecordSize    := fBlobBufOffset + fBlobBufSize;
  GetMem(fFilterBuffer, fRecordSize);
  fCurrent := 0;
  BookmarkSize := sizeOf(Pointer);
  FIsTableOpen := True;
end;

procedure TWcsMemDataSet.InternalPost;
var
  LDSRecInfo : PLDSRecInfo;
  i  : integer;
  ListObj  : TObject;
  FieldBuf : Pointer;
  NullFlag : Boolean;
  TempBuff : PChar;
  OldListCount, OldIdx, ExtIdx: Longint;
begin
// All Edits are performed on "fEditObj"
// If we had full record buffering we would write the fields
// here from the active buffer and there would be no "fEditObj"

// For proper index support ascertain the correct insert position
// in the internal list using compare routines....
  LDSRecInfo  := RecInfoFromBuffer(GetActiveRecBuffer);
  OldListCount := ListCount;
  if (State = dsInsert) then
  begin
    // Append or insert as first
    if (fCurrent = 0) or (LDSRecInfo^.BookmarkFlag = bfEOF) or eof then
    begin
      ExtIdx := OldListCount;
      ListInsert(fEditObject,ExtIdx);
      fCurrent := fIdxList.Add(TObject(ExtIdx))+1;
    end else begin
      ExtIdx := Longint(LDSRecInfo^.Bookmark);
      ListInsert(fEditObject,ExtIdx);
      fIdxList.Insert(fCurrent-1,TObject(ExtIdx));
  // May not be where we expect if the external list insists
  // on keeping things in some sort of order.
  // But in any case, if the external list INSERTED, we must reindex....
      if (ExtIdx < OldListCount) then
        IndexReset(ExtIdx, OldListCount-1, 1);
    end;
  end else begin
    ExtIdx := Longint(LDSRecInfo^.Bookmark);
    OldIdx := ExtIdx;
    ListSetObject(fEditObject, ExtIdx);
    if ExtIdx > OldIdx then
      IndexReset(OLdIdx+1, ExtIdx, -1)
    else
      if ExtIdx < OldIdx then
        IndexReset(ExtIdx, OldIdx-1, 1);

  end;
  fIdxList[fCurrent-1] := TObject(ExtIdx);
  fEditObject := Nil;
  fEditBuff   := Nil;
  LDSRecInfo^.IntIdx := fCurrent;
  LDSRecInfo^.Bookmark := Pointer(ExtIdx);
end;

procedure TWcsMemDataSet.InternalSetToRecord(Buffer: PChar);
begin
  fCurrent := fIdxList.IndexOf(PLDSRecInfo(Buffer+fRecBufSize).Bookmark)+1;
end;

function TWcsMemDataSet.IsCursorOpen: Boolean;
begin
  Result := FIsTableOpen;
end;

procedure TWcsMemDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  PLDSRecInfo(Buffer+fRecBufSize).Bookmark := TObject(Data^);
end;

procedure TWcsMemDataSet.SetBookmarkFlag(Buffer: PChar;
  Value: TBookmarkFlag);
begin
  PLDSRecInfo(Buffer+fRecBufSize).BookmarkFlag := Value;
end;

procedure TWcsMemDataSet.SetRecNo(Value: Integer);
begin
  if fCurrent <> Value then
  begin
    if (Value < 1) or (Value > FIdxList.Count) then
      raise Exception.Create ('SetRecNo: out of range');
    fCurrent := Value;
    Resync([rmExact, rmCenter]);
    DoAfterScroll;
  end;
end;

constructor TWcsMemDataSet.Create(Owner: TComponent);
begin
  inherited;
  fIdxList := TList.Create;
end;

destructor TWcsMemDataSet.Destroy;
begin
  inherited;
  FIdxList.Free;
end;

procedure TWcsMemDataSet.SetExtList(const Value: TObject);
begin
  CheckInactive;
  fExtList := Value;
end;

function TWcsMemDataSet.GetActiveRecBuffer : PChar;
begin
  Result := Nil;
  case State of
    dsBrowse: if not IsEmpty then
                Result := ActiveBuffer;
    dsEdit, dsInsert: Result := ActiveBuffer;
    //dsSetKey: Result := PChar(fKeyBuffer);
    dsCalcFields: Result := CalcBuffer;
    dsFilter: Result := fFilterBuffer;
  end;
end;

function TWcsMemDataSet.RecInfoFromBuffer(Buffer: PChar) : PLDSRecInfo;
begin
  if Buffer <> Nil then
    Result := PLDSRecInfo(Buffer+fRecBufSize)
  else
    Result := Nil;
end;

function TWcsMemDataSet.RecObjFromBuffer(Buffer: PChar): TObject;
var
  ExtIdx: Longint;
begin
  if Buffer <> Nil then
  begin
    if Buffer = fEditBuff then
      Result := fEditObject
    else begin
      ExtIdx := Longint(PLDSRecInfo(Buffer+fRecBufSize)^.Bookmark);
      Result := ListGetObject(ExtIdx);
    end;
  end else
    Result := Nil;
end;

procedure TWcsMemDataSet.ReadListData;
var
  i : Longint;
begin
  fIdxList.Clear;
  for i := 0 to ListCount - 1 do
    fIdxList.Add(Pointer(i));
end;

function TWcsMemDataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  RecBuf: PChar;
  RecInfo: PLDSRecInfo;
  ExtIdx: Longint;
  ListObj: TObject;
begin
  RecBuf := GetActiveRecBuffer;
  if RecBuf <> Nil then
  begin
    RecInfo := RecInfoFromBuffer(RecBuf);
    if Field.FieldNo > 0 then
    begin
      ExtIdx:= Longint(RecInfo.Bookmark);
      ListObj := RecObjFromBuffer(RecBuf);
      Result := ListGetField(ListObj, ExtIdx, Field, Buffer);
    end else begin // calcfield
      Inc(RecBuf, FRecordSize + Field.Offset);
      Result := Boolean(RecBuf[0]);
      if Result and (Buffer <> nil) then
        Move(RecBuf[1], Buffer^, Field.DataSize);
    end;
  end else
    Result := False;
end;

procedure TWcsMemDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  RecBuf: PChar;
  ListObj : TObject;
begin
  RecBuf := GetActiveRecBuffer;
  if RecBuf <> Nil then
  begin
    if Field.FieldNo > 0 then
    begin
      if State = dsCalcFields then DatabaseError(SNotEditing, Self);
      if ReadOnly and not (State in [dsSetKey, dsFilter]) then
        DatabaseErrorFmt(SFieldReadOnly, [Field.DisplayName]);
      ListObj := RecObjFromBuffer(RecBuf);
      ListSetField(ListObj, Field, Buffer);
    end else begin // Calculated
      Inc(RecBuf, fRecordSize + Field.Offset);
      Boolean(RecBuf[0]) := LongBool(Buffer);
      if Boolean(RecBuf[0]) then Move(Buffer^, RecBuf[1], Field.DataSize);
    end;
    if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
      DataEvent(deFieldChange, Longint(Field));
  end else
    raise exception.Create('Invalid List Index or Record Buffer in SetFieldData');
end;

procedure TWcsMemDataSet.SortByFunc(const SortFunc: TWcsMemDataSetCompare);
var
  BM : TBookmark;
begin
  fSortList := fIdxList.List;
  if (fSortList <> Nil) and (fIdxList.Count > 1) then
  begin
    DisableControls;
    try
      if Active then
        BM := GetBookMark;
      fQSLevel := 0;
      QuickSort(0,fIdxList.Count-1,SortFunc);
      if Active then
      begin
        Refresh;
        if BM <> Nil then
        begin
          GotoBookmark(BM);
          FreeBookmark(BM);
        end;
      end;
    finally
      enablecontrols;
    end;
  end;
end;

procedure TWcsMemDataSet.QuickSort(L, R: Integer;
  const SCompare: TWcsMemDataSetCompare);
var
  I, J: Integer;
  P, T: Pointer;
begin
  inc(fQSLevel);
  if fQSLevel > 100 then
    if R < L then Exit;
  repeat
    I := L;
    J := R;
    P := ListGetObject(integer(fSortList^[(L + R) shr 1]));
    repeat
      while SCompare(ListGetObject(integer(fSortList^[I])), P) < 0 do Inc(I);
      while SCompare(ListGetObject(integer(fSortList^[J])), P) > 0 do Dec(J);
      if I <= J then
      begin
        T := fSortList^[I];
        fSortList^[I] := fSortList^[J];
        fSortList^[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSort(L, J, SCompare);
    L := I;
  until I >= R;
  dec(fQSLevel);
end;

procedure TWcsMemDataSet.Sort;
begin
  if Assigned(fOnSortCompare) then
    SortByFunc(fOnSortCompare);
end;

procedure TWcsMemDataSet.ListIndexSort;
var
  BM : TBookmark;
begin
  DisableControls;
  try
    if Active then
      BM := GetBookMark;
    ReadListData;
    if Active then
    begin
      Refresh;
      if BM <> Nil then
      begin
        GotoBookmark(BM);
        FreeBookmark(BM);
      end;
    end;
  finally
    enablecontrols;
  end;
end;

procedure TWcsMemDataSet.DoAfterScroll;
begin
  fCurrent := PLDSRecInfo(ActiveBuffer+fRecBufSize).IntIdx;
end;

function TWcsMemDataSet.CompareBookmarks(Bookmark1,
  Bookmark2: TBookmark): Integer;
var
  I1, I2 : integer;
begin
  I1 := PLongint(Bookmark1)^;
  I2 := PLongint(Bookmark2)^;
  if I1 > I2 then  Result := 1
  else if I2 > I1 then Result := -1
  else Result := 0;
end;

procedure TWcsMemDataSet.InternalInitFieldDefs;
begin
  ListSetupFieldDefs;
end;

procedure TWcsMemDataSet.ListSetupFieldDefs;
begin
// by default we do nothing
end;

procedure TWcsMemDataSet.ListSetupFields;
begin
// by default we do nothing
end;

procedure TWcsMemDataSet.ListDataRead;
begin
// DO something AFTER reading the list but before acknowledging Active
end;

function TWcsMemDataSet.CreateBlobStream(Field: TField;
  Mode: TBlobStreamMode): TStream;
begin
  Result := TWcsMemDataSetBlobStream.Create(Field, Mode);
end;

procedure TWcsMemDataSet.ListDataFetch;
begin
// Do nothing
end;

procedure TWcsMemDataSet.SetReadOnly(const Value: Boolean);
begin
  CheckInactive;
  fReadOnly := Value;
end;

{ TWcsMemListDataSet }

function TWcsMemListDataSet.getDBList: TList;
begin
  Result := TList(fExtList);
end;

procedure TWcsMemListDataSet.SetDBList(const Value: TList);
begin
  SetExtList(Value);
end;

function TWcsMemListDataSet.ListCount: Integer;
begin
  if fExtList = Nil then
    Result := 0
  else
    Result := TList(fExtList).Count;
end;

function TWcsMemListDataSet.ListGetObject(index: integer): TObject;
begin
  Result := TList(fExtList)[Index];
end;

procedure TWcsMemListDataSet.ListDelete(const DelIdx: integer);
begin
  ListFreeElement(TList(fExtList)[DelIdx]);
  TList(fExtList).Delete(DelIdx);
end;

procedure TWcsMemListDataSet.ListFreeElement(const Element: TObject);
begin
  if Element <> Nil then Element.Free;
end;

procedure TWcsMemListDataSet.ListInsert(const NewObj: TObject;
  var NewIdx: integer);
begin
  TList(fExtList).Insert(NewIdx, NewObj);
end;

procedure TWcsMemListDataSet.ListSetObject(NewObj: TObject;
  var NewIdx: Integer);
begin
  ListFreeElement(TList(fExtList)[NewIdx]);
  TList(fExtList)[NewIdx] := NewObj;
end;


{ TWcsMemDataSetBlobStream }

constructor TWcsMemDataSetBlobStream.Create(AField : TField; AMode: TBlobStreamMode);
var
  RecBuf: PChar;
  bSize : integer;
begin
  inherited Create;
  fMode := AMode;
  fField := AField;
  with TWcsMemDataSet(fField.Dataset) do
  begin
    if fMode in [bmWrite, bmReadWrite] then
    begin
      if fField.ReadOnly then DatabaseErrorFmt(SFieldReadOnly,
        [fField.DisplayName], fField.DataSet);
      if not (fField.DataSet.State in [dsEdit, dsInsert]) then
        DatabaseError(SNotEditing, fField.DataSet);
    end;
    RecBuf := GetActiveRecBuffer;
    fObject := RecObjFromBuffer(RecBuf);
    if fMode in [bmRead, bmReadWrite] then
    begin
      if ListGetBlob(fObject, fField, Nil, bSize) then
        if bSize > 0 then
        begin
          SetSize(bSize);
          ListGetBlob(fObject, fField, Memory, bSize);
        end;
    end;
  end;
end;

destructor TWcsMemDataSetBlobStream.Destroy;
begin
  if fMode in [bmWrite, bmReadWrite] then
    TWcsMemDataSet(fField.Dataset).ListSetBlob(fObject, fField, Memory, Size);
  inherited;
end;



{ TWcsArticleDataSet }

constructor TWcsArticleDataSet.Create(Owner: TComponent);
begin
  inherited;
  fArticles := TList.Create;
// Enable the standard Count, GetObject
  ExtList := fArticles;
end;

destructor TWcsArticleDataSet.Destroy;
begin
  inherited;
  ClearArticles;
  fArticles.Free;
end;

procedure TWcsArticleDataSet.ClearArticles;
var
  i : integer;
begin
  for i := 0 to fArticles.Count - 1 do
    ListFreeElement(fArticles[i]);
  fArticles.Clear;
  fLastId := 0;
end;

procedure TWcsArticleDataSet.ListSetupFieldDefs;
begin
  FieldDefs.Clear;
  FieldDefs.Add('Sequence',ftInteger,0,False);
  FieldDefs[0].Attributes := [DB.faReadonly];
  FieldDefs.Add('Id',ftInteger,0,False);
  FieldDefs[1].Attributes := [DB.faReadonly];
  FieldDefs.Add('Title',ftString, 40);
  FieldDefs.Add('Text',ftMemo);
end;

function TWcsArticleDataSet.ListGetField(ListObject : TObject;
  ListIndex: integer; Field: TField; FieldBuf: Pointer): Boolean;
var
  S : String;
  L : integer;
begin
  Result := ListObject <> Nil;
  if Result then
  with TArticleItem(ListObject) do
  case Field.FieldNo of
    1 : if FieldBuf <> Nil then PLongint(FieldBuf)^ := ListIndex;
    2 : if FieldBuf <> Nil then PLongint(FieldBuf)^ := ArticleID;
    3 : begin
          Result := Length(ArticleTitle) > 0;
          if FieldBuf <> Nil then StrPCopy(PChar(FieldBuf),ArticleTitle);
        end;
    4 : Result := Length(ArticleText) > 0; // don't return data
  else Result := False;
  end;
end;

procedure TWcsArticleDataSet.ListSetField(
  ListObject : TObject; Field: TField; FieldBuf: Pointer);
var
  L : integer;
begin
  if ListObject <> Nil then
  with TArticleItem(ListObject) do
  case Field.FieldNo of
    3 : if FieldBuf = Nil then
          fArticleTitle := ''
        else begin
          fArticleTitle := PChar(FieldBuf);
        end;
    4 : if FieldBuf = Nil then
          fArticleText := '';
  end;
end;

procedure TWcsArticleDataSet.AlphaSort;
begin
  SortByFunc(DBTitleCompare);
end;

procedure TWcsArticleDataSet.IDSort;
begin
  SortByFunc(DBIDCompare);
end;

function TWcsArticleDataSet.DBTitleCompare(Object1,
  Object2 : TObject): Integer;
begin
  Result := CompareText(TArticleItem(Object1).fArticleTitle,
                        TArticleItem(Object2).fArticleTitle);
end;

function TWcsArticleDataSet.DBIdCompare(Object1,
  Object2 : TObject): Integer;
var
  I1, I2 : DWord;
begin
  I1 := TArticleItem(Object1).fArticleID;
  I2 := TArticleItem(Object2).fArticleID;
  if I1 > I2 then  Result := 1
  else if I2 > I1 then Result := -1
  else Result := 0;
end;

procedure TWcsArticleDataSet.ListInsert(const NewObj : TObject; var NewIdx: integer);
begin
  inherited;
  inc(fLastId);
  TArticleItem(NewObj).fArticleId := fLastId;
end;

function TWcsArticleDataSet.ListCreateElement(const Element: TObject): TObject;
begin
  Result := TArticleItem.Create;
  with TArticleItem(Result) do
  if Element <> Nil then
  begin
    fArticleID := TArticleItem(Element).fArticleID;
    fArticleTitle := TArticleItem(Element).fArticleTitle;
    fArticleText := TArticleItem(Element).fArticleText;
  end else
    TArticleItem(Result).fArticleID := fLastId+1;
end;

function TWcsArticleDataSet.ListGetBlob(const ListObject: TObject;
  const Field: TField; const Buffer: Pointer;
  var BufLen: integer): Boolean;
begin
  if (ListObject <> Nil) and (Field.FieldNo = 4) then
  with TArticleItem(ListObject) do
  begin
    Result := Length(fArticleText) > 0;
    if Result then
    begin
      BufLen := Length(fArticleText);
      if Buffer <> Nil then
        Move(fArticleText[1], Buffer^, BufLen);
    end;
  end else
    Result := False;
end;

procedure TWcsArticleDataSet.ListSetBlob(const ListObject: TObject;
  const Field: TField; const Buffer: Pointer; const BufLen: Integer);
begin
  if Field.FieldNo = 4 then
  with TArticleItem(ListObject) do
  begin
    SetLength(fArticleText, BufLen);
    if BufLen > 0 then Move(Buffer^, fArticleText[1], BufLen);
  end;
end;

procedure TWcsArticleDataSet.SaveToStream(const Stream: TStream);
var
  i, L : Longint;
  S : String;
begin
  Stream.Write(fLastID, 4);
  L := fArticles.Count;
  Stream.Write(L, 4);
  for i := 0 to fArticles.Count-1 do
  with TArticleItem(fArticles[i]) do
  begin
    Stream.Write(fArticleId, 4);
    L := Length(fArticleTitle);
    Stream.Write(L, 4);
    if L > 0 then Stream.Write(fArticleTitle[1], L);
    L := Length(fArticleText);
    Stream.Write(L, 4);
    if L > 0 then Stream.Write(fArticleText[1], L);
  end;
end;

procedure TWcsArticleDataSet.LoadFromStream(const Stream: TStream);
var
  i, L,C : Longint;
  S : String;
  ArticleItem : TArticleItem;
begin
  ClearArticles;
  Stream.Read(fLastID, 4);
  Stream.Read(C, 4);
  for i := 0 to C-1 do
  begin
    ArticleItem := TArticleItem.Create;
    fArticles.Add(ArticleItem);
    with ArticleItem do
    begin
      Stream.Read(fArticleId, 4);
      Stream.Read(L, 4);
      SetLength(fArticleTitle,L);
      if L > 0 then Stream.Read(fArticleTitle[1], L);
      Stream.Read(L, 4);
      SetLength(fArticleText,L);
      if L > 0 then Stream.Read(fArticleText[1], L);
    end;
  end;
end;

end.

